
/*****************************************************************************
 *                                                                           *
 *  PMB source code, MPI-1 part                           ---/---            *
 *                                                          / /              *
 *  Copyright (c) PALLAS GmbH 1998                         / / /             *
 *                                                        / / / /            *
 *                                                         / / /             *
 *  Authors: Hans-Joachim Plum , Gero Ritzenhoefer          / / PALLAS       *
 *  Revision: 2.0                                         ---/---            *
 *  Date: 1998/03/10                                                         *
 *                                                                           *
 *****************************************************************************/

/****************************************************************************
Initialization of an array of "benchmark descriptors" of type Bench
(see Benchmark.h)
*****************************************************************************/


#include <stdio.h>

#include "declare.h"
#include "Benchmark.h"
#include "Comments.h"

/* PROTOTYPES OF BENCHMARK FUNCTIONS */
extern void  PingPong  (struct comm_info*,int,int,double*);

/* NAMES OF BENCHMARKS (DEFAULT CASE)*/
char *def_cases[] = {
   "PingPong" 
  ,"PingPong" 
  ,NULL
};

char * NIL_COMMENT[] ={NULL};

void Construct_BList(struct Bench** P_BList, char* name)

/**************************************************************************
             VARIABLE |       TYPE        |   MEANING
------------------------------------------------------------
Input      : 
             name     | char*             | name of benchmark to insert
                      |                   | ="ALL": take all default benchmarks
                      |                   |
Output     : P_BList  | struct Bench**    | First call: 
                      |                   | *P_Blist is allocated
                      |                   | current benchmark is inserted
                      |                   |
                      |                   | Following calls
                      |                   | current benchmark is inserted
                      |                   |
------------------------------------------------------------
**************************************************************************/

{

 static int Ndeflt = -1 ;
 static int n_cases;

 struct Bench* Bmark;

 int i, plc;


 if( Ndeflt < 0 )
   {
  Ndeflt=-1;
  while( def_cases[++Ndeflt] );
  *P_BList = (struct Bench*) malloc((1+Ndeflt)*sizeof(struct Bench));
  if( *P_BList == NULL ){ Err_Hand(0, APPL_ERR_MALLOC);}

  n_cases=0;
   }

 if( !strcmp(name,"ALL") )
  {
  for( plc=0; plc<Ndeflt; plc++ )
    Construct_BList(P_BList, def_cases[plc]);

  return;
  }

 Bmark = (*P_BList)+n_cases;

 if( (Bmark->name =(char *)malloc(sizeof(char) * (1+strlen(name)) ))
     == NULL ) Err_Hand(0, APPL_ERR_MALLOC);

   for( i=0; i<strlen(name); i++ )
   if ( name[i] >= 'A' && name[i] <= 'Z' )
        Bmark->name[i] = name[i] - ('A'-'a');
   else Bmark->name[i] = name[i];

 Bmark->name[strlen(name)] = '\0';
 
 Bmark->bench_comments = &NIL_COMMENT[0];
 Bmark->scale_time = 1.0;
 Bmark->scale_bw   = 1.0;


 if (!strcmp(Bmark->name,"pingpong"))
	{ 
          strcpy(Bmark->name,"PingPong");
          Bmark->Benchmark = PingPong;
          Bmark->bench_comments = &PingPong_cmt[0];
          Bmark->type = SingleTransfer;
          Bmark->scale_time = 0.5;
	}
 else 
        {
          strcpy(Bmark->name,name);
          Bmark->type = BTYPE_INVALID;
        }

  n_cases++;
  (*P_BList)[n_cases].name=NULL;

}

void Destruct_BList(struct Bench ** P_BList)
{
/****************************************************************
Freeing of the Benchmark list
*****************************************************************/
int i;
i=0;
while( (*P_BList)[i].name )
  {
  free ((*P_BList)[i++].name);
  }
free(*P_BList);
}

void print_BList(struct comm_info * c_info, struct Bench *BList)
/****************************************************************
Display of the Benchmark list
*****************************************************************/
{
int j, ninvalid;
char*nn,*cmt;

if( General_cmt[0] != NULL )
        {
        fprintf(unit,"# Remarks on the current Version:\n\n");
        j=0;
        while ( (nn=General_cmt[j++]) )
           fprintf(unit,"# %s\n",nn);
        }


j=0; ninvalid=0;

while( BList[j].name )
  {
   if( BList[j].type == BTYPE_INVALID )
     {
     ninvalid++;
     if( ninvalid==1 )
          fprintf(unit,"\n# Attention, invalid benchmark name(s):\n");
     
          fprintf(unit,"# %s\n",BList[j].name);
          free (BList[j].name);
          BList[j].name = (char*)malloc(sizeof(char));
          strcpy(BList[j].name,"");
     }
  j++;
  }


if( ninvalid < j)
{

fprintf(unit,"\n# List of Benchmarks to run:\n\n");

j=0;
while((nn=BList[j].name))
  {
  if( BList[j].type != BTYPE_INVALID )
  {
  if( c_info->group_mode >= 0 )
  fprintf(unit,"# (Multi-)%s\n",nn);
  else
  fprintf(unit,"# %s\n",nn);

  if    ( *(BList[j].bench_comments) )
      fprintf(unit,"#     Comments on this Benchmark:\n");
  while ( (cmt = *(BList[j].bench_comments++)) )
      fprintf(unit,"#     %s\n",cmt);

  }
  j++;
  }
}

}
