
/*****************************************************************************
 *                                                                           *
 *  PMB source code, MPI-1 part                           ---/---            *
 *                                                          / /              *
 *  Copyright (c) PALLAS GmbH 1998                         / / /             *
 *                                                        / / / /            *
 *                                                         / / /             *
 *  Authors: Hans-Joachim Plum , Gero Ritzenhoefer          / / PALLAS       *
 *  Revision: 2.0                                         ---/---            *
 *  Date: 1998/03/10                                                         *
 *                                                                           *
 *****************************************************************************/
/*-----------------------------------------------------------------
Several output functions

Functions implemented in this module: 

void Output         (struct comm_info* c_info,struct Bench* Bmark,
	             int NP, int header,int size,int n_sample,double *time)
void display_times  (struct Bench* Bmark,
                     double* tlist,struct comm_info* c_info,
                     int group,int n_sample,int size,int edit_type)
void print_headlines(struct comm_info* c_info,int tmp_NP, 
                     char* cases,int li_len,char* h_line)
void Show_selections
                    (struct comm_info* c_info, struct Bench* BList);
void Show_procids
                    (struct comm_info* c_info);

void make_line      (int li_len)
void edit_format    (int n_ints ,int n_floats)
int Str_Atoi        (char s[])

                     
-----------------------------------------------------------------*/
#include <string.h>
#include "declare.h"
#include "Benchmark.h"

/* EXTERNAL PROTOTYPES */
extern void Err_Hand(int,int);
extern void print_BList();

/* LOCAL   PROTOTYPES */
void Output         (struct comm_info* c_info,struct Bench* Bmark,
	             int NP, int header,int size,int n_sample,double *time);
void display_times  (struct Bench* Bmark,
                     double* tlist,struct comm_info* c_info,
                     int group,int n_sample,int size,int edit_type);

void accumulate_eff_bandwidth  (struct Bench* Bmark,
                     double* tlist,struct comm_info* c_info,
                     int group,int n_sample,int size,int edit_type);

void print_headlines(struct comm_info* c_info,int tmp_NP, 
                     char* cases,int li_len,char* h_line);
void Show_selections
                    (struct comm_info* c_info, struct Bench* BList);
void Show_procids
                    (struct comm_info* c_info);

void make_line      (int li_len);
void edit_format    (int n_ints ,int n_floats);
int Str_Atoi        (char s[]);

/* max time over all groups and min bandwidth */

static double group_tmax     = 0;
static double group_bw_min   = 0;
static double eff_bandwidth  = 0;
static int    counter        = 0;

/*****************************************************************/
void Output(struct comm_info* c_info,struct Bench* Bmark,
	    int tmp_NP, int header,int size,int n_sample,double *time)
/*****************************************************************/

/*-----------------------------------------------------------------
             VARIABLE |       TYPE        |   MEANING
-------------------------------------------------------------------
Input      :  c_info  | struct comm_info* | see comm_info.h 
              Bmark   | struct Bench*     | current enchmark
              tmp_NP  | int               | number of nodes
	      header  | int               | first call or not (header flag)
              size    | int               | message length in byte
	      n_sample| int               | repetition count
	      time    | double*           | time of measurement
	              |                   |
Output     :          |                   |
                      |                   |
In/Out     :  -       | -                 | -
                      |                   |  
-------------------------------------------------------------------
-------------------------------------------------------------------
Description: Output of results (header and bare data of measurement)
-----------------------------------------------------------------*/
{
 static double* all_times=NULL;
  static int     NP_flag=-1;

  double scaled_time;
  
  int DO_OUT;
  int GROUP_OUT;
  int i,k,i_gr;
  int li_len;
  int ierr;
  int edit_type;

  MPI_Status stat;
  
  DO_OUT    = (c_info->w_rank  == 0 );
  GROUP_OUT = (c_info->group_mode > 0 );

  if (DO_OUT) 
    {
      if(!all_times)
	{
	  all_times = (double*) malloc( c_info->w_num_procs * sizeof(double) );
	  if( !all_times ) Err_Hand( 0, APPL_ERR_MALLOC );
	}
#ifdef CHECK
      if(!all_defect)
	{
	  all_defect = (double*) calloc( c_info->w_num_procs , sizeof(double)); 
	  if( !all_defect ) Err_Hand( 0, APPL_ERR_MALLOC );	
	}
#endif  	  
    }

/* Scale the timings */
  scaled_time = *time * SCALE*Bmark->scale_time;

/* collect all times  */
  ierr=MPI_Gather(&scaled_time,1,MPI_DOUBLE,all_times,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
  MPI_ERRHAND(ierr);

#ifdef CHECK      
/* collect all defects */	      
  ierr=MPI_Gather(&defect,1,MPI_DOUBLE,all_defect,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
  MPI_ERRHAND(ierr);
#endif
  if( DO_OUT )
    {
      if ( Bmark->type == SingleTransfer && c_info->group_mode != 0 )
           edit_type=0;
      else if ( Bmark->type == MultiTransfer || Bmark->type == SingleTransfer )
           edit_type=1;
      else if (Bmark->type == Global && strcmp(Bmark->name,"Barrier"))
           edit_type=2;
      else if ( !strcmp(Bmark->name,"Barrier") )
           edit_type=3;

      if( header )
	{
          if( GROUP_OUT ) {strcpy(aux_string,"%Group") ; li_len=1;}
          else            {strcpy(aux_string,"");  li_len=0;}
	  if ( edit_type == 0 )
	    {
	      li_len+=4;
	      strcat(aux_string,"%#bytes%#repetitions%t[usec]%Mbytes/sec%");
	    }
	  else if ( edit_type == 1 )
	    {
	      li_len+=6;
	      strcat(aux_string,
		     "%#bytes%#repetitions%t_min[usec]%t_max[usec]%t_avg[usec]%Mbytes/sec%");
	    }
	  else if ( edit_type == 2 )
	    {
	      li_len+=5;
	      strcat(aux_string,
		     "%#bytes%#repetitions%t_min[usec]%t_max[usec]%t_avg[usec]%");
	    }
	  else
	    {
	      li_len+=4;
	      strcat(aux_string,
		     "%#repetitions%t_min[usec]%t_max[usec]%t_avg[usec]%");
	    }
#ifdef CHECK
          if( strcmp(Bmark->name,"Barrier") )
          {
	  li_len+=1;
	  strcat(aux_string,"%defects%");
          }
#endif
	  print_headlines(c_info,tmp_NP,Bmark->name,li_len,aux_string);
	}     

      if( GROUP_OUT )
	{

	  group_tmax = 0.0;

	  k=0;
	  for( i_gr=0; i_gr<c_info->n_groups; i_gr++ )
	    {
	      if(i_gr == 0) fprintf(unit,"\n");
	      
	      display_times(Bmark, all_times, c_info, i_gr, n_sample, size, edit_type);
    
	      k+= c_info->g_sizes[i_gr];
	    } 
	}
      else
	display_times(Bmark, all_times, c_info,  0, n_sample, size, edit_type);

      accumulate_eff_bandwidth(Bmark, all_times, c_info,  0
                                                  , n_sample, size, edit_type);


    } 
}

/*****************************************************************/
void display_times( struct Bench* Bmark,
                    double* tlist,struct comm_info* c_info,int group,
		    int n_sample,int size,int edit_type)
/*****************************************************************/

/*-------------------------------------------------------------------
             VARIABLE |       TYPE        |   MEANING
-------------------------------------------------------------------
Input      :  Bmark   | struct Bench*     | Current benchmark 
              tlist   | double*           | list of timings
	      c_info  | struct comm_info* | current comm. structure
	      group   | int               | number of group
	      n_sample| int               | repetition count
              size    | int               | message length in byte
                      |                   |
Output     :          |                   |
                      |                   |
In/Out     :  -       | -                 | -
                      |                   |  
-------------------------------------------------------------------
-------------------------------------------------------------------
Description: Output of results (header and bare data of measurement)
-----------------------------------------------------------------*/
{
  double tmax, tmin, tav, defect;
  int i, ip;
  
  if( c_info->g_sizes[group]<= 0 ) return;

  if( c_info->group_mode > 0)
  {
  
  i =0;
  ip=0;
  while( i<group )
    {
      ip+= c_info->g_sizes[i++];
    }
 
    tmin = tlist[ip];
    tmax = 0.;
    tav  = 0.;
    defect = 0.;
    for(i=0; i<c_info->g_sizes[group]; i++)
	{
	  tmax  =  max(tmax,tlist[ip+i]);
	  tmin  =  min(tmin,tlist[ip+i]);
	  tav  += tlist[ip+i];
#ifdef CHECK 
          defect = max ( defect, all_defect[c_info->g_ranks[ip+i]] );
#endif 
	}
    tav /= c_info->g_sizes[group];

    /* max over all groups -- EFF_BW */

    if( tmax >= group_tmax ) group_tmax = tmax;

    }
    else
    {
    ip = 0;
    for( i=0; i<c_info->n_groups; i++ ) ip += c_info->g_sizes[i];
    tmin = tlist[0];
    tmax = 0.;
    tav  = 0.;
    defect = 0.;
    for(i=0; i<ip; i++)
	{
	  tmax  =  max(tmax,tlist[i]);
	  tmin  =  min(tmin,tlist[i]);
	  tav  += tlist[i];
#ifdef CHECK 
          defect = max ( defect, all_defect[i] );
#endif 
	}
    tav /= ip;
    }

    ip=0;
    if(c_info->group_mode > 0) 
      {
	edit_format(1,0);
	sprintf(aux_string,format,group);
	ip=strlen(aux_string);
      }

    if( edit_type == 0 )
      { 
	edit_format(2,2);
	sprintf(aux_string+ip,format,size,n_sample,
		tmax,(Bmark->scale_bw*SCALE*1.e-6)*size/tmax);
      }
    else if ( edit_type == 1 )
      {
	edit_format(2,4);
	sprintf(aux_string+ip,format,size,n_sample,tmin,tmax,
		tav,(Bmark->scale_bw*SCALE*1.e-6)*size/tmax);
      }
    else if ( edit_type == 2 )
      {
	edit_format(2,3);
	sprintf(aux_string+ip,format,size,n_sample,tmin,tmax,tav);
      }
    else if ( edit_type == 3 )
      {
	edit_format(1,3);
	sprintf(aux_string+ip,format,n_sample,tmin,tmax,tav);
      }
#ifdef CHECK 
    if ( edit_type != 3 )
    {
    edit_format(0,1);
    ip=strlen(aux_string);
    sprintf(aux_string+ip,format,1000*all_defect[0]);
    }
#endif
    fprintf(unit,"%s\n",aux_string);
    fflush(unit);
  
}


/*****************************************************************/

void accumulate_eff_bandwidth( struct Bench* Bmark,
                    double* tlist,struct comm_info* c_info,int group,
		    int n_sample,int size,int edit_type)
/*****************************************************************

 description: accumulates effective bandwidth and counter



*******************************************************************/

{
  double tmax, tmin, tav, defect;
  int i, ip;
  

  group_bw_min = (Bmark->scale_bw*SCALE*1.e-6)*size
                              *(c_info->w_num_procs)/(2.0*group_tmax);


  if( size > 0 ) 
    {
      eff_bandwidth += group_bw_min;
      counter       += 1;
    } 
  
}

void Output_eff_bandwidth(struct comm_info* c_info)
/**************************************************************

 description: prints out effective bandwidth on rank == 0 PE.


***************************************************************/
{

#include "mpi.h"


  MPI_Barrier( MPI_COMM_WORLD);

  if( c_info->w_rank == 0 )
    {

      printf("\n\n***********************************************************\n\n");

      printf("Running on %d PEs\n\n", c_info->w_num_procs );

      printf("sampling from 2^0 to 2^%d  bytes\n\n", counter/2 - 1 );

      printf("Effective Bandwidth: %6.2lf [MB/sec] \n", eff_bandwidth/((double) counter ));

      printf("\n\n***********************************************************\n\n");

    }

}


/************************************************************************/
void Show_selections(struct comm_info* c_info, struct Bench* BList)
/************************************************************************

-----------------------------------------------------------------------------
             VARIABLE |       TYPE        |   MEANING
-------------------------------------------------------------------------
Input      : c_info   | struct comm_info* | see comm_info.h 
             BList    | struct Bench*     | list of benchmarks to run
                      |                   |
Output     :     -    | -                 | - 
                      |                   |  
In/Out     :     -    | -                 | -
                      |                   |  
------------------------------------------------------------------------
------------------------------------------------------------------------
Description: Output of basic parameters
------------------------------------------------------------------------
*/
{
  int ip, i,k;
  char*p;
  if(c_info->w_rank == 0 )
    {
      General_Info();
      fprintf(unit,"#\n");
      fprintf(unit,"# Minimal message length         :   %d\n",0);
      fprintf(unit,"# Maximal message length in bytes:   %d\n",1<<MAXMSGLOG);
      fprintf(unit,"#\n");
      fprintf(unit,"# MPI_Datatype                   :   MPI_BYTE \n");
      fprintf(unit,"# MPI_Datatype for reduction     :   MPI_FLOAT\n");
      fprintf(unit,"# MPI_Op                         :   MPI_SUM  \n");
      fprintf(unit,"#\n");

      print_BList(c_info, BList);

    }
}

/****************************************************************************/
void Show_procids(struct comm_info* c_info)
/****************************************************************************

-----------------------------------------------------------------------------
             VARIABLE |       TYPE        |   MEANING
-----------------------------------------------------------------------------
Input      : c_info   | struct comm_info* | see comm_info.h 
                      |                   |
Output     :     -    | -                 | - 
                      |                   |  
In/Out     :     -    | -                 | -
                      |                   |  
---------------------------------------------------------------------------
---------------------------------------------------------------------------
Description: Output of number of processes and number of 
             active groups (incl. nodes)
---------------------------------------------------------------------------
*/
{
  int ip, i, k, idle;
  char**p;
  
  if( c_info->w_rank == 0 )
    {
      if(c_info->n_groups == 1)
        {
	fprintf(unit,"\n# ( #processes = %d ) \n",c_info->num_procs);
        idle = c_info->w_num_procs-c_info->num_procs;
        }
      if(c_info->n_groups != 1)
	{
	  fprintf(unit,"\n# ( %d groups of %d processes each running simultaneous ) \n",
		  c_info->n_groups,c_info->num_procs); 
	  ip = 0;
	  for(i=0; i<c_info->n_groups; i++)
	    {
	      fprintf(unit,"# Group %4d                     :",i);
	      for(k=0; k<c_info->g_sizes[i]; k++) 
		fprintf(unit," %d", c_info->g_ranks[ip+k]); 
	      ip += c_info->g_sizes[i];
	      fprintf(unit,"\n");
	    }
          idle = c_info->w_num_procs - ip;
	}
      if( idle )
        {
        if( idle == 1 )
	fprintf(unit,"# ( %d additional process waiting in MPI_Barrier)\n",idle);
        else
	fprintf(unit,"# ( %d additional processes waiting in MPI_Barrier)\n",idle);
        }
    }

}


/*****************************************************************/
void print_headlines(struct comm_info* c_info,int tmp_NP,char* cases,
		     int li_len,char* h_line)
/*****************************************************************/
{
  char* help;
  char* token;

  fprintf(unit,"\n");            /* FOR GNUPLOT: CURVE SEPERATOR  */

  make_line(li_len);
  if( c_info->n_groups > 1) 
  fprintf(unit,"# Benchmarking Multi-%s ",cases);
  else
  fprintf(unit,"# Benchmarking %s ",cases);
  Show_procids(c_info); 

  make_line(li_len);
  
  help=h_line;
  while(token = strtok(help, "%") )
    {
      sprintf(format,"%%%ds",ow_format);
      fprintf(unit,format,token);
      help = NULL;
    }
  fprintf(unit,"\n");
}

/*****************************************************************/
void edit_format(int n_ints , int n_floats)
/*****************************************************************/
{
  int ip,i;
  
  ip=0;
  for(i=1 ; i<=n_ints; i++)
    {    
      sprintf(&(format[ip]),"%%%dd",ow_format); 
      ip=strlen(format);
    }
  for(i=1 ; i<=n_floats ; i++)
    {
      sprintf(&(format[ip]),"%%%d.2f",ow_format);
      ip=strlen(format);
    }
}

/***************************************************************************/
void make_line(int li_len)
/***************************************************************************

-----------------------------------------------------------------------------
             VARIABLE |       TYPE        |   MEANING
----------------------------------------------------------------------------
Input      :  li_len  | int               | line length
-----------------------------------------------------------------------------
----------------------------------------------------------------------------
Description: print a line of li_len many "-"
---------------------------------------------------------------------------*/
{
  int i;
  char* char_line = "-";
  fprintf(unit,"#");
  for( i=1;i<li_len*ow_format; i++ )
    {
      fprintf(unit,"%s",char_line);
    }
  fprintf(unit,"\n");	
}


/***************************************************************************/
int Str_Atoi(char s[])
/***************************************************************************

-----------------------------------------------------------------------------
             VARIABLE |       TYPE        |   MEANING
----------------------------------------------------------------------------
Input      :     s    | char              | asci string
                      |                   |
Return     :          | int               | numerical value of digits in string
                      |                   |  
-----------------------------------------------------------------------------
----------------------------------------------------------------------------
Description: Converts ascii string to int up to char symbol ' '
---------------------------------------------------------------------------*/
{
  int i,n,len ;
  n=0;
  len=strlen(s);
  for(i=0; s[i] >= '0' && s[i] <= '9' || s[i]==' '; ++i)
    {
      if (s[i]!=' ') 
	{
	  n=10*n+(s[i]-'0');
	}
    }
  if (i<len)
    {
      n=-1;
    }
  return n;
}

